# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2017-2024, by Samuel Williams.
# Copyright, 2020, by Salim Semaoune.

require_relative "async/version"
require_relative "async/reactor"

require_relative "kernel/async"
require_relative "kernel/sync"

# Asynchronous programming framework.
module Async
end
